"use strict";

const tls   = require('tls');
const net   = require('net');
const debug = require('debug');


const Client  = require('./client.js');
const merge   = require('mout/object/merge');
const forIn   = require('mout/object/forIn');
const Events  = require('eventemitter-co');
const defer      = require('nyks/promise/defer');

const EVENT_SOMETHING_APPEND = 'change_append';

const log = {
  info  : debug('ubk:server'),
  error : debug('ubk:server'),
};


const evtmsk = function(ns, cmd) {
  return `_${ns}:${cmd}`;
};

class Server extends Events {

  constructor(options) {
    super();
    this._clientsList = {};
    this._rpcs        = {};
    this._clientHeartBeat = null;

    this.options = merge({
      'secured'       : false,
      'server_port'   : 8000,
      'socket_port'   : 8001,
      'heartbeat_interval' : 1000 * 20,
      'broadcasting_registration'  : true,
      'tls_options' : {
        'requestCert' : true,
        'rejectUnauthorized' : true,
        'key' :  null,
        'cert' : null,
        'ca' : [null]
      }
    }, options);

    this.heartbeat = this.heartbeat.bind(this);
    this.new_tcp_client = this.new_tcp_client.bind(this);
    this.new_websocket_client = this.new_websocket_client.bind(this);
    this.get_client = this.get_client.bind(this);
    this.register_client = this.register_client.bind(this);
    this.lost_client = this.lost_client.bind(this);
    this.call = this.call.bind(this);

    if(this.options.secured) {
      this.tcp_server = tls.createServer(this.options.tls_options, this.new_tcp_client);
    } else {
      this.tcp_server = net.createServer(this.new_tcp_client);
    }

    //... what else ?
    this.register_rpc('base', 'ping', () => 'pong');
    this.register_cmd('base', 'register', this.register_client);

    this.register_cmd('base', 'register_sub_client', async(client, query) => {
      var sub_client_registrationargs = query.args;
      var error;
      var response;
      try {
        await this.register_sub_client(client, sub_client_registrationargs);
      } catch(err) {
        log.error(err);
        error = err;
      }
      client.respond(query, response, error);
    });

    this.register_cmd('base', 'unregister_sub_client', async(client, query) => {
      var sub_client_key    = query.args.client_key;
      var error;
      var response;
      try {
        this.unregister_sub_client(client, sub_client_key);
      } catch(err) {
        log.error(err);
        error = err;
      }
      client.respond(query, response, error);
    });
  }

  async register_sub_client(client, sub_client_registrationargs) {
    var sub_client_key    = sub_client_registrationargs.client_key;
    var client_capability = sub_client_registrationargs.client_capability;
    var all_sub_client = this.get_all_sub_client();
    if(all_sub_client[sub_client_key])
      throw `Client '${sub_client_key}' already exists, sorry`;
    var validated_data = await this.validate_sub_client(sub_client_key, client_capability);
    var sub_client = client.add_sub_client(sub_client_key);
    sub_client.export_json = () => {
      var status     = sub_client_registrationargs.export_json;
      status.uptime  = Math.floor((Date.now() / 1000 - status.registration_time));
      return status;
    };
    this.emit('register_sub_client', sub_client, validated_data).catch(log.error);
  }

  unregister_sub_client(client, sub_client_key) {
    var sub_client = client._sub_clients[sub_client_key];
    if(!sub_client)
      throw `Client '${sub_client_key}' already unregistred`;
    client.remove_sub_client(sub_client.client_key);
    this.emit('unregister_sub_client', sub_client).catch(log.error);
  }

  async validate_sub_client() {
    return true; // to be overrided
  }

  get_client(client_key) {
    return this._clientsList[client_key];
  }

  get_all_sub_client() {
    var all_sub_client = {};
    forIn(this._clientsList, (client) => {
      all_sub_client = merge(all_sub_client, client._sub_clients);
    });
    return all_sub_client;
  }

  start() { /*chain*/
    var args = [].slice.apply(arguments);
    var chain = args.shift() || Function.prototype;

    var defered = defer();
    var server_port = this.options.server_port;

    this._clientHeartBeat = setInterval(this.heartbeat,  this.options.heartbeat_interval);

    log.info("Server is in %s mode", this.options.secured ? "SECURED" : "NON SECURED");

    this.tcp_server.listen({port : server_port, host : '0.0.0.0'}, (err) => {
      this.options.server_port  =  this.tcp_server.address().port;
      log.info("Started TCP server for clients on port %d", this.options.server_port);
      defered.chain(err, this.options.server_port);
      chain();
    });

    return defered;
  }

  heartbeat() {

    forIn(this._clientsList, (client) => {
      // Check failures
      if(client.ping_failure) {
        log.info("client %s failed ping challenge, assume disconnected", client.client_key);
        return client.disconnect();
      }

      // Send ping
      client.ping_failure = true;
      client.send('base', 'ping').then(function(response) {
        client.ping_failure = !(response == 'pong');
      });
    });
  }


  // Build new client from tcp stream
  new_tcp_client(stream) {
    log.info("Incoming tcp stream");
    var client = new Client('tcp', stream);
    client.once('received_cmd', this.register_client);
  }

  // Build new client from web socket stream
  new_websocket_client(stream) {
    log.info("Incoming ws stream");
    var client = new Client('ws', stream);
    client.once('received_cmd', this.register_client);
  }


  async register_client(client, query) {
    try {
      var args = query.args;
      //can only register once...
      if(query.ns != 'base' || query.cmd != 'register')
        throw `Un-expected registration query`;

      if(client.client_key)
        throw `Already registered client '${client.client_key}'`;

      client.client_key = args.client_key;
      // Check SSL client cert matches
      var exp = client.export_json();

      if(exp.secured && exp.name != client.client_key)
        throw `The cert '${exp.name}' does NOT match the given id '${client.client_key}'`;

      if(!client.client_key)
        throw `No id for client to register`;

      // Avoid conflicts
      if(this._clientsList[client.client_key])
        throw `Client '${client.client_key}' already exists, sorry`;

      try {
        for(var sub in args.sub_Clients_list || [])
          await this.register_sub_client(client, sub);
      } catch(error) {
        log.error('cant register subClient', error);
      }

    } catch(err) {
      if(typeof query == 'object')
        client.respond(query, null, err);
      return client.disconnect();
    }

    // Save client
    this._clientsList[client.client_key] = client;

    client.respond(query, 'ok');
    client.once('disconnected', this.lost_client);
    client.on('received_cmd', this._onMessage.bind(this));

    // THAT'S GREAT, LET'S NOTIFY EVERYBOOOOODYYYY
    client.emit('registered', args).catch(log.error);
    this.emit('registered_device', client, args).catch(log.error);
    if(this.options.broadcasting_registration)
      this.broadcast('base', 'registered_client', client.export_json());
  }


  lost_client(client) {
    // Remove from list
    log.info("Lost client", client.client_key);

    forIn(client._sub_clients, (sub_client) => {
      try {
        this.unregister_sub_client(client, sub_client.client_key);
      } catch(err) {
        log.error(err);
      }
    });

    delete this._clientsList[client.client_key];

    this.emit('unregistered_device', client).catch(log.error);
    if(this.options.broadcasting_registration)
      this.broadcast('base', 'unregistered_client', {client_key : client.client_key });
  }

  unregister_cmd(ns, cmd) {
    this.off(evtmsk(ns, cmd));
  }


  register_cmd(ns, cmd, callback, ctx) {
    this.off(evtmsk(ns, cmd));
    this.on(evtmsk(ns, cmd), callback, ctx);
  }


  async call(ns, cmd) {
    var args = [].slice.call(arguments, 2);
    var proc = this._rpcs[evtmsk(ns, cmd, 'rpc')];
    if(!proc)
      throw "Invalid rpc command";
    return await proc.callback.apply(proc.ctx || this, args);
  }

  register_rpc(ns, cmd, callback, ctx) {

    this._rpcs[evtmsk(ns, cmd, 'rpc')] = {callback, ctx};

    this.register_cmd(ns, cmd, async function(client, query) {
      var response;
      var error;
      try {
        var args = [query.args].concat(query.xargs || []);
        response = await callback.apply(this, args);
      } catch(err) { error = '' + err; }

      client.respond(query, response, error);
    }, ctx);
  }

  async _onMessage(client, data) {
    var target = data.ns;
    if(typeof target == 'string') {
      let tmp = target.split(':'); //legacy ns:device_key syntax
      target = { ns : tmp[0], client_key : tmp[1] };
    }

    if(target.client_key) { //proxy
      log.info("proxy %s from %s to %s", data, client.client_key, target.client_key);
      var remote = this._clientsList[target.client_key];
      var response;
      var error;

      if(!remote)
        remote = this.get_all_sub_client()[target.client_key];
      try {
        if(!remote)
          throw `Bad client '${target.client_key}'`;
        response = await remote.send.apply(remote, [target.ns, data.cmd, data.args].concat(data.xargs));
      } catch(err) {
        error = err;
      }
      return client.respond(data, response, error);
    }

    var ns  = target.ns;
    var cmd = data.cmd;
    this.emit(evtmsk(ns, cmd), client, data)
      .then(() => {
        this.emit(EVENT_SOMETHING_APPEND, ns, cmd).catch(log.error);
      })
      .catch(log.error);
  }


  broadcast (ns, cmd, payload) {
    var args = arguments;
    log.info("BROADCASTING", ns, cmd);

    forIn(this._clientsList, function(client) {
      client.signal.apply(client, args);
    });

    this.emit(`${ns}:${cmd}`, payload).catch(log.error);
  }

}


module.exports = Server;
